################################################################################################
# Library of functions to provide example code for integration with shot managers
################################################################################################
import hiero
import os
import time
import shot_manager_xlsx
from PySide2 import QtGui
from PySide2 import QtWidgets

def __assign_task_tags(item, task_list):
    """Assign a tag to track_item based on task"""   
    # map task names to some default icons (coudl se custom ones here as well)
    icons = {'comp':'icons:TagNuke.png',
                     'textures':'icons:TagMari.png',
                     'light':'icons:TagKatana.png',
                     'anim':'icons:TagModo.png',
                     'fx':'icons:TagFurnace.png',
                     'track':'icons:TagCameraTracker.png'}
    # assign a tag for each task. If no icon is defined for a task name use the "ready to start" icon
    for task in task_list:
        # only ad a tag if there is none for this task yet
        if not [tag for tag in item.tags() if tag.name() == task]:
            item.addTag(hiero.core.Tag(task, icons.get(task, 'icons:status/TagReadyToStart.png')))

def __assign_notes_tag(item, vfx_req):
    """If vfx requirements were found add a tag and add the string to it as a note"""
    tag_name = 'VFX requirements'
    if vfx_req:
        existing_tag = [tag for tag in item.tags() if tag.name() == tag_name]
        if not existing_tag:
            # create a new tag
            tag = hiero.core.Tag(tag_name, 'icons:TagComment.png')
            item.addTag(tag)
        else:
            # re-use existing tag
            tag = existing_tag[0]
        tag.setNote(vfx_req)

def __create_new_seq_item(project, name, fps=25, start_tc='01:00:00:00'):
    """Create a sequence wth the given name and add it to the project bin"""
    # create a new sequence with the given name
    seq = hiero.core.Sequence(name)
    
    # set sequence basics
    seq.setFramerate(fps)
    seq.setTimecodeStart(start_tc, False) # assume no drop frame
    
    # wrap it into a bin item and add it to the main bin
    project.clipsBin().addItem(hiero.core.BinItem(seq))
    return seq

def __get_project():
    """Get the last opened project"""
    return hiero.core.projects()[-1]

def __get_task_colour(task_list):
    """Define colours for certain task combinations"""

    if len(task_list) == 1 and task_list[0] == 'comp':
        # make comp only shots green
        return QtGui.QColor(0,60,30)

    if 'fx' in task_list:
        # make shot red if it requires a fx task
        return QtGui.QColor(80,40,30)

def __get_seq_by_name(name):
    """Find and return the sequence in the current priject with the given name.
    If it doesn't exist create and return a new sequence with that name.
    """
    project = __get_project()
    for seq in project.sequences():
        # return the existing sequence with the given name
        if seq.name() == name:
            return seq
    # if we get this far no sequence has been found with the given name,
    # so create a new one and return it
    return __create_new_seq_item(project, name)

def __task_string_to_list(task_string):
    """Turn the return value from the spreadsheet into a tidy list"""
    if not task_string:
        # no tasks were assigned in the spreadsheet
        return []
    return [s.strip() for s in task_string.split(',')]


def get_db_handle(silent=False):
    """Retrieve the value set in the custom "Database Settings" panel, then return a handle to for it's I/O.
    If the path is not valid pop up a warning.
    if silent is True no panel will pop up so we can use this in spawned threads, e.g. the export processor.
    """
    xlsx_path = hiero.ui.database_settings_panel.line_edit_widget.text()
    if os.path.isfile(xlsx_path) and os.path.splitext(xlsx_path)[-1] == ('.xlsx'):
        # path s valid
        return shot_manager_xlsx.DB_IO(xlsx_path)

    if not silent:
        # path is not valid, let the user know
        msgBox = QtWidgets.QMessageBox()
        msgBox.setText('File Path in Database Settings panel ins invalid.\n Please update it to point to a valid xlsx file.')
        msgBox.exec_()
        hiero.ui.windowManager().showWindow(hiero.ui.database_settings_panel)
    
def get_track_items():
    """Get the selection in the currently active timeline.
    If nothing is selected return all track items in the sequence.
    Only consider video track items.
    """
    # get current selection
    seq = hiero.ui.getTimelineEditor(hiero.ui.activeSequence())
    track_item_list = seq.selection()
    if not track_item_list:
        # if there is no selection collect all track items
        seq.selectAll()
        track_item_list = seq.selection()
        seq.selectNone()
    # filter out video track items only
    return [ti for ti in track_item_list if isinstance(ti, hiero.core.TrackItem)]

def collect_shot_data():
    """Collect file name, and timecodes from selected track items.
    Return a list of dictionaries containing the respective data"""
    # collect data form the selected track items
    shot_data = []
    for ti in get_track_items():
        seq = ti.sequence()
        # add the sequnece start timecode as offset to each track items's timecode
        timeline_in_frames = ti.timelineIn()
        timeline_out_frames = ti.timelineOut()
        # times are in frames, so convert them to strings representing timecodes based on the sequence's frame rate
        timeline_in_tc = hiero.core.Timecode.timeToString(timeline_in_frames, seq.framerate(), hiero.core.Timecode.kDisplayTimecode)
        timeline_out_tc = hiero.core.Timecode.timeToString(timeline_out_frames, seq.framerate(), hiero.core.Timecode.kDisplayTimecode)
        file_path = ti.source().mediaSource().fileinfos()[0].filename() # only use first file fragment for simplicity
        # collect the data
        shot_data.append({'File Name':    ti.source().name(),
                          'File Path':    file_path,
                          'Shot Code':    ti.name(),
                          'Sequence':     ti.sequence().name(),
                          'Timeline In':  timeline_in_tc,
                          'Timeline Out': timeline_out_tc,
                          'Handles Head': ti.handleInLength(),
                          'Handles Tail': ti.handleOutLength()
                         })
    return shot_data

def create_track_from_data(shots_data):
    """Create a sequence in Hiero based on provided shot data"""
    # create a time stamp to label tracks and bins
    time_stamp = time.ctime()
    
    # create a new bin with the time stamp
    project = __get_project()
    bin = hiero.core.Bin(time_stamp)
    project.clipsBin().addItem(bin)

    new_seqs = []
    # iterate over each item in the list which represents data for a given shot
    for shot_data in shots_data:   
        # create a new media item for the plate specified in the shot data
        media_source = hiero.core.MediaSource(shot_data['File Path'])
    
        # create a clip item with the new media
        clip = hiero.core.Clip(media_source)
    
        # create bin item and colour code it based on task assignment in spreadsheet
        bin_item = hiero.core.BinItem(clip)
        
        # add item to bin
        bin.addItem(bin_item)

        # get the target sequence
        seq = __get_seq_by_name(shot_data['Sequence'])     

        # convert timecodes from spreadsheet to frames based on sequence frame rate
        seq_in = hiero.core.Timecode.stringToTime(shot_data['Timeline In'], seq.framerate(), hiero.core.Timecode.kDisplayTimecode)[0]
        seq_out = hiero.core.Timecode.stringToTime(shot_data['Timeline Out'], seq.framerate(), hiero.core.Timecode.kDisplayTimecode)[0]

        # get the tasks from shot data
        task_list = __task_string_to_list(shot_data['Tasks'])
            
        # add clip to the sequence. This creates and links audio track items as required
        track_items = seq.addClip(clip, seq_in)

        for track_item in track_items:
            track_item.setName(shot_data['Shot Code'])

            # set timings according to the spreadsheet
            track_item.setTimes(seq_in,
                                seq_out,
                                shot_data['Handles Head'],
                                clip.duration() - shot_data['Handles Tail'] - 1)
        

            # do stuff that is specific to video tracks only
            if isinstance(track_item, hiero.core.TrackItem):
                # time stamp the new track
                track_item.parentTrack().setName(time_stamp)
    
                # add tags for vfx description if there are any, but only to video track items
                __assign_notes_tag(track_item, shot_data['VFX requirements'])
    
                # if tasks are assigned in teh spreadsheet add tags for them
                __assign_task_tags(track_item, task_list)
    
        # colour code the bin item based on task assignment in spreadsheet
        bin_item.setColor(__get_task_colour(task_list))
        
        # collect affected sequence items so we can open them when all is done
        new_seqs.append(seq.binItem())

    for seq_item in list(set(new_seqs)):
        # update internal state and send any signals that Hiero may require
        seq_item.activeItem().editFinished()

    # open the new sequence in the timeline
    hiero.ui.openInTimeline(list(set(new_seqs)))

def update_track_items(shots_data):
    """Use shot_data to update selected track items with tags and colours.
    shots_data is a list of dictionaries as returned by shot_manager_xlsx.DB_IO.read_shot_data():
    """
    # create a dictionary that uses timeline in and our as keys, so it's to look up for a given track item
    shots_by_tc = {}
    seq = hiero.ui.activeSequence()
    for shot_data in shots_data:
        # turn timecode into frames
        time_in = hiero.core.Timecode.stringToTime(shot_data.pop('Timeline In'), seq.framerate(), hiero.core.Timecode.kDisplayTimecode, seq.timecodeStart())[0]
        time_out = hiero.core.Timecode.stringToTime(shot_data.pop('Timeline Out'), seq.framerate(), hiero.core.Timecode.kDisplayTimecode, seq.timecodeStart())[0]
        shots_by_tc[(time_in, time_out)] = shot_data
    
    # loop over selected track items and look up the corresponding shot data
    for track_item in get_track_items():
        try:
            shot_data = shots_by_tc[(track_item.timelineIn() - seq.timecodeStart(), track_item.timelineOut() - seq.timecodeStart())]
        except KeyError:
            # no timecode match found, so skip this track item
            print("couldn't find {} in spreadsheet".format((track_item.timelineIn(), track_item.timelineOut())))
            continue
    
        # assign shot name
        if shot_data['Shot Code']:
            track_item.setName(shot_data['Shot Code'])

        # add tags for vfx description if there are any
        __assign_notes_tag(track_item, shot_data['VFX requirements'])

        # deal with tasks
        task_list = __task_string_to_list(shot_data['Tasks'])

        # if tasks are assigned in the spreadsheet add tags for them
        __assign_task_tags(track_item, task_list)

        # colour code the bin item based on task assignment in spreadsheet
        track_item.source().binItem().setColor(__get_task_colour(task_list))
