######################################################
# Example code as a generic database interface.
# This is likely to be replaced with inhouse database APIs
# and only serves to help illustrate exampled for integration with Hiero/Nuke Studio
# 
# Required libraries: openpyxl 2.5.5 or later (included), pillow/pil
# To install PIL/pillow run:
# pip install pillow
#
# NOTE:
#     Make sure to include the path to PIL libraries in your PYTHONPATH to be available in Hiero
#     e.g.: export PYTHONPATH="/usr/local/lib/python2.7/site-packages/"
#           or update load_lib.py accordingly
#
######################################################
import os
import openpyxl
import time

class DB_IO(object):
    """Read from and write to an xlsx file.
    This is a placeholder for any shot managers that may be in use such as ShotGun, FTrack etc.
    """

    def __init__(self, xlsx_path):
        """Instantiate with the path to the xlsx file and load up the workbook"""
        self.xlsx_path = xlsx_path
        self.book = openpyxl.load_workbook(self.xlsx_path)
    
    def __get_headers(self, sheet):
        for row in sheet.rows:
            return [cell.value for cell in row]
        
    def read_shot_data(self, sheet_name):
        """Return a list of dictonaries representing shot data (each item in the list is a shot).
        E.g.: [{u'Plate': u'path/to/frames/seq1/seq1.####.exr', u'Tasks': u'anim,light,comp', u'Handles Head': 8,
               u'Sequence': u'A', u'Shot Code': u'A_010', u'Handles Tail': 8, u'Seq Timecode In': '00:00:01:03', u'Seq Timecode Out': '00:00:02:13'},
               ...]
        """
        sheet = self.book[sheet_name]
        shot_data = []
        for i, row in enumerate(sheet.rows):
            if i == 0:
                # collect header data on the first iteration
                header_data = [cell.value for cell in row]
                continue
            shot_data.append(dict(zip(header_data, [cell.value for cell in row])))
        return shot_data
    
    
    def write_shot_data(self, shot_data, sheet_name):
        """Write filenames and timecodes for shots to the xlsx file.
        shot_data is a list of dictionaries which hold the data we want to write.
        """
        sheet = self.book[sheet_name]
        header_data = self.__get_headers(sheet)

        for i, shot in enumerate(shot_data):
            sheet.append([shot.get(key, '') for key in header_data])
        self.book.save(self.xlsx_path)

    def write_version_data(self, version_data, sheet_name):
        """Write version data to the xlsx file.
        version_data is a dictionary which holds the data we want to write.
        """
        sheet = self.book[sheet_name]
        header_data = self.__get_headers(sheet)

        # add current time as time stamp
        version_data['Submitted'] = time.ctime()
        sheet.append([version_data.get(key, None) for key in header_data])

        if version_data['Thumbnail'] and os.path.isfile(version_data['Thumbnail']):
            # if a valid path to a thumbanli is provided add it to the spreadsheet
            img = openpyxl.drawing.image.Image(version_data['Thumbnail'])
            # adjust height for last row to make space for the thumbnail
            sheet.row_dimensions[sheet.max_row].height = img.height * .75
            # add thumbnail to "B" column
            sheet.add_image(img, 'B{}'.format(sheet.max_row))
        self.book.save(self.xlsx_path)

if __name__ == '__main__':
    db_handle = DB_IO('/ohufx/consulting/Foundry/ShotManagerDemo/src/ShotData.xlsx')
    
    version_data = {'Version Name':'test',
                     'Thumbnail':'/Users/frank/Desktop/dontBackup/review/2018_08_01/MFB_A_020_comp_thumb.jpg',
                     'Frame Range': '1-777',
                     'Path To Frames': None,
                     'Path To Mov': '/Users/frank/Desktop/dontBackup/review/2018_08_01/MFB_A_020_comp.mov'
                     }                 

    db_handle.write_version_data(version_data, 'playlist')
